#[allow(unused_imports)]
use chrono::{FixedOffset, NaiveDate, NaiveDateTime};

#[cfg(feature = "mal")]
use myanimelist::{auth::Auth, objects::AnimeNode, AccessToken, RefreshToken};
use serde::{Deserialize, Serialize};
use std::{fmt::Debug, str::FromStr};

fn default_filename() -> String {
    return "all.ics".to_string();
}

fn default_length() -> i64 {
    return 25;
}

fn default_shift() -> String {
    return "+0100".into();
}

#[derive(Deserialize, Serialize, Debug, Clone, Eq, PartialOrd, Ord)]
pub(crate) struct Anime {
    pub(crate) name: String,
    pub(crate) episodes: u64,
    pub(crate) start: chrono::NaiveDateTime,
    #[serde(default = "default_length")]
    pub(crate) length: i64,
    #[serde(default = "default_shift")]
    shift: String,
    pub(crate) url: Option<String>,
    #[serde(default)]
    pub(crate) pause: Vec<chrono::NaiveDate>,
    #[serde(default)]
    pub(crate) description: Option<String>,
}

impl Anime {
    pub(crate) fn shift(&self) -> FixedOffset {
        return FixedOffset::from_str(
            self.shift
                .clone()
                .as_str(),
        )
        .unwrap();
    }
}

#[cfg(feature = "mal")]
impl From<AnimeNode> for Anime {
    fn from(value: AnimeNode) -> Self {
        let offset: FixedOffset = FixedOffset::east_opt(9 * 60 * 60).unwrap();
        let broadcast = value
            .broadcast
            .unwrap();
        let start_date = value
            .start_date
            .unwrap();
        let mut title = value.title;
        if let Some(alt_titles) = value.alternative_titles {
            if let Some(en_title) = alt_titles.en {
                title = en_title;
            }
        }
        Self {
            name: title,
            episodes: match value.num_episodes {
                None => 13,
                Some(0) => 13,
                Some(i) => i.into(),
            },
            start: NaiveDateTime::new(
                NaiveDate::from_ymd_opt(
                    start_date
                        .year
                        .into(),
                    start_date
                        .month
                        .unwrap()
                        .into(),
                    start_date
                        .day
                        .unwrap()
                        .into(),
                )
                .unwrap(),
                broadcast.start_time,
            )
            .checked_sub_offset(offset)
            .unwrap(),
            length: match value.average_episode_duration {
                Some(i) => (i as i64).div_euclid(60),
                None => 25,
            },
            shift: "+0100".into(),
            url: Some(format!("https://myanimelist.net/anime/{}", value.id)),
            pause: Vec::new(),
            description: value.synopsis,
        }
    }
}

impl PartialEq for Anime {
    fn eq(&self, other: &Self) -> bool {
        self.name == other.name
    }
}

#[cfg(feature = "mal")]
#[derive(Deserialize, Serialize, Debug, Clone)]
pub(crate) struct MALAccesskey {
    pub(crate) access_token: AccessToken,
    pub(crate) refresh_token: RefreshToken,
    pub(crate) access_expiry_time: u64,
    pub(crate) refresh_expiry_time: u64,
}

#[cfg(feature = "mal")]
impl MALAccesskey {
    pub(crate) fn load(&self, auth: &Auth) {
        auth.set_access_token_unchecked(
            self.access_token
                .clone(),
        );
        auth.set_expires_at_unchecked(self.access_expiry_time);
        auth.set_refresh_token_unchecked(
            self.refresh_token
                .clone(),
        );
        auth.set_refresh_expires_at_unchecked(self.refresh_expiry_time);
    }
}

#[derive(Deserialize, Serialize, Debug)]
pub(crate) struct Config {
    pub(crate) outputdir: String,
    #[serde(default = "default_filename")]
    pub(crate) complete_ics: String,

    #[cfg(feature = "mal")]
    pub(crate) mal_accesskey: Option<MALAccesskey>,

    #[serde(default)]
    pub(crate) anime: Vec<Anime>,
}

#[derive(clap::Parser, Debug)]
#[command(author, version, about, disable_help_subcommand = true)]
pub(crate) struct Commandline {
    #[arg(short, long, default_value = "/etc/animecalendar.toml")]
    pub(crate) config: String,
    #[command(subcommand)]
    pub(crate) subcommand: Option<Subcommand>,
}

#[derive(clap::Subcommand, Debug, Clone, Copy)]
pub(crate) enum Subcommand {
    #[cfg(feature = "mal")]
    #[command(name="sync", visible_aliases=["sync-mal","mal"] ,about="Adds the missing shows from MyAnimelist")]
    SyncMAL,
}
