use chrono::Duration;
use clap::{crate_name, Parser};
use icalendar::{Calendar, Component, Event, EventLike, Property};
use lazy_static::lazy_static;
use log::{self};
#[cfg(feature = "mal")]
use malclient::update_config;
use muddy::muddy_init;
use std::{
    collections::HashMap,
    fs,
    io::{self, ErrorKind, Read, Seek, Write},
    os::unix::fs::OpenOptionsExt,
    path::Path,
};
use stderrlog::{new as errlog, ColorChoice, Timestamp};
use systemd_journal_logger::JournalLog;
use tokio::runtime::LocalRuntime;

mod config;
#[cfg(feature = "mal")]
mod malclient;

lazy_static! {
    static ref CONFIG_READOPTIONS: fs::OpenOptions = fs::OpenOptions::new()
        .write(false)
        .read(true)
        .create(false)
        .to_owned();
    static ref CALENDAR_WRITEOPTIONS: fs::OpenOptions = fs::OpenOptions::new()
        .write(true)
        .create(true)
        .truncate(true)
        .read(false)
        .mode(0o644)
        .to_owned();
    static ref CONFIG_WRITEOPTIONS: fs::OpenOptions = fs::OpenOptions::new()
        .write(true)
        .create(true)
        .truncate(false)
        .read(true)
        .mode(0o644)
        .to_owned();
}

muddy_init!("embed");

fn slugify(input: String) -> String {
    input
        .to_lowercase()
        .replace(" ", "-")
        .replace(",", "")
}

fn gen_cal(name: &str, description: &str) -> Calendar {
    Calendar::new()
        .name(name)
        .description(description)
        .ttl(&Duration::days(1))
        .append_property(Property::new("CUTYPE", "OTHER"))
        .append_property(Property::new("FBTYPE", "BUSY-UNAVAILABLE"))
        .done()
}

fn write_calendars(config: config::Config) {
    match fs::DirBuilder::new()
        .recursive(true)
        .create(
            config
                .outputdir
                .clone(),
        ) {
        Ok(()) => {},
        Err(err) => match err.kind() {
            ErrorKind::AlreadyExists => {},
            error => {
                log::error!("Failed to create directory: {}", error);
                return;
            },
        },
    }
    let binding = config
        .outputdir
        .clone();
    let p = Path::new(&binding);
    let mut allcal = gen_cal("Anime Calendar", "Calendar for all configured animes");
    let mut itemcal: Calendar;
    let mut formatargs = HashMap::<String, String>::with_capacity(2);
    formatargs.insert("name".to_string(), "".to_string());
    formatargs.insert("episode".to_string(), "".to_string());
    for anime in config
        .anime
        .iter()
    {
        itemcal = gen_cal(format!("Calendar for {}", anime.name).as_str(), format!("Calendar for {}", anime.name).as_str());
        formatargs.insert(
            "name".to_string(),
            anime
                .name
                .clone(),
        );
        let mut lastdate = anime
            .start
            .checked_add_offset(anime.shift())
            .unwrap();
        let mut pause = anime
            .pause
            .clone();
        pause.sort();
        for episode in 1..(anime.episodes + 1) {
            formatargs.insert("episode".to_string(), episode.to_string());
            let mut start = if episode == 1 {
                lastdate
            } else {
                lastdate
                    .checked_add_days(chrono::Days::new(7))
                    .unwrap()
            };
            for breakdate in pause.iter() {
                if start.date() == *breakdate {
                    start = start
                        .checked_add_days(chrono::Days::new(7))
                        .unwrap();
                }
            }
            lastdate = start;
            let mut item = Event::new()
                .starts(start)
                .ends(
                    start
                        .checked_add_signed(Duration::minutes(anime.length))
                        .unwrap(),
                )
                .summary(
                    anime
                        .name
                        .clone()
                        .as_str(),
                )
                .alarm(icalendar::Alarm::display(&anime.name, Duration::seconds(0)))
                .uid(slugify(format!("{}-{}", anime.name, episode)).as_str())
                .done();
            match anime
                .description
                .clone()
            {
                None => {},
                Some(desc) => {
                    item.description(desc.as_str());
                },
            }
            match anime
                .url
                .clone()
            {
                None => {},
                Some(url) => {
                    let url = strfmt::strfmt(url.as_str(), &formatargs.clone()).unwrap();
                    item.url(url.as_str())
                        .location(url.as_str());
                },
            }
            allcal.push(item.clone());
            itemcal.push(item.done());
        }
        let mut itemics = CALENDAR_WRITEOPTIONS
            .open(p.join(format!("{}.ics", anime.name.clone())))
            .unwrap();
        write!(itemics, "{}", itemcal).unwrap();
    }
    let mut allics = CALENDAR_WRITEOPTIONS
        .open(p.join(format!("{}", config.complete_ics)))
        .unwrap();
    write!(allics, "{}", allcal).unwrap();
}

fn main() {
    match JournalLog::new() {
        Ok(log) => {
            log.add_extra_field("RUST_PACKAGE", crate_name!())
                .install()
                .unwrap();
        },
        Err(_) => {
            io::stderr()
                .write(b"Failed to setup logging")
                .unwrap();
            errlog()
                .color(ColorChoice::Auto)
                .quiet(false)
                .show_level(true)
                .show_module_names(true)
                .timestamp(Timestamp::Second)
                .init()
                .unwrap();
            return;
        },
    }
    log::set_max_level(log::LevelFilter::Trace);
    let args = config::Commandline::parse();
    let configfileresult = match args.subcommand {
        None => CONFIG_READOPTIONS.open(args.config),
        Some(_) => CONFIG_WRITEOPTIONS.open(args.config),
    };
    let mut configfile = match configfileresult {
        Ok(c) => c,
        Err(e) => {
            log::error!("Error opening the configuration: {}", e.to_string());
            return;
        },
    };
    let mut cfgstring = String::new();
    match configfile.read_to_string(&mut cfgstring) {
        Ok(_) => {},
        Err(error) => {
            log::error!("Failed to read config file: {}", error);
            return;
        },
    }
    #[allow(unused_mut)]
    let mut config: config::Config = toml::from_str(&cfgstring).expect("Expected valid toml");
    #[allow(unused_variables)]
    let runtime = LocalRuntime::new().expect("Failed to get LocalRuntime");
    #[allow(unreachable_patterns)]
    match args.subcommand {
        None => {
            log::info!("ICal creator started up");
            log::debug!("{:?}", config);
            write_calendars(config);
            return;
        },
        #[cfg(feature = "mal")]
        Some(config::Subcommand::SyncMAL) => {
            log::info!("Updating the list");
            let _guard = runtime.enter();
            runtime.block_on(update_config(&mut config));
        },
        Some(_) => {},
    }
    match toml::to_string_pretty(&config) {
        Ok(configstring) => {
            let _ = configfile.seek(io::SeekFrom::Start(0));
            let _ = configfile.set_len(0);
            let _ = configfile.write(configstring.as_bytes());
            let _ = configfile.sync_all();
        },
        Err(_) => todo!(),
    };
}
