use std::{
    error::Error,
    net::{Ipv4Addr, SocketAddr},
    time::Duration,
};

use crate::config::{self, Anime, MALAccesskey};
use lazy_static::lazy_static;
use log::{debug, error, info};
use muddy::{muddy_all, muddy_init};
use myanimelist::{
    self,
    auth::Auth,
    objects::{Username, WatchStatus},
    AuthorizationCode, ClientId, ClientSecret, CsrfToken, MalClient, MalClientError, RedirectUrl, Scope,
};
use std::sync::Mutex;
use tokio::{task, time::sleep};

#[cfg(feature = "hyper")]
mod hyperserver;
#[cfg(feature = "hyper")]
use hyperserver::hyperserver as serverfuture;

const SCHEME: &str = "http";
const HOST: &str = "localhost";
const PORT: u16 = 8080;
const LOCALHOSTV4: SocketAddr = SocketAddr::new(std::net::IpAddr::V4(Ipv4Addr::new(127, 0, 0, 1)), PORT);
const LIMIT: u16 = 100;

lazy_static! {
    static ref CALLBACKRESULT: Mutex<AuthorizationCode> = Mutex::new(AuthorizationCode::new(String::new()));
    static ref CSRFTOKEN: Mutex<CsrfToken> = Mutex::new(CsrfToken::new(String::new()));
}

muddy_init!();

muddy_all! {
    static CLIENTID: &str = env!("MAL_CLIENTID");
    static CLIENTSECRET: &str = env!("MAL_CLIENTSECRET");
}

async fn callback(_url: reqwest::Url, token: CsrfToken) -> Result<(AuthorizationCode, CsrfToken), Box<dyn Error>> {
    {
        let mut csrftoken = CSRFTOKEN
            .lock()
            .unwrap();
        *csrftoken = CsrfToken::new(
            token
                .secret()
                .clone(),
        );
    }
    println!("Please visit {} to authenticate", _url.as_str());
    loop {
        sleep(Duration::from_nanos(100)).await;
        let lock = match CALLBACKRESULT.lock() {
            Ok(cg) => cg,
            Err(error) => {
                log::error!("Failed to get call guard: {error}");
                return Err(Box::new(error));
            },
        };
        if lock
            .secret()
            .len()
            != 0
        {
            return Ok((
                AuthorizationCode::new(
                    lock.secret()
                        .clone(),
                ),
                token,
            ));
        }
    }
}

async fn get_animes(client: &MalClient, animelist: &mut Vec<Anime>, status: WatchStatus) {
    let mut offset: u64 = 0;
    loop {
        debug!("Getting animes, round {}", offset / LIMIT as u64);
        match client
            .user_animelist()
            .get()
            .user_name(Username::Me)
            .offset(offset)
            .limit(LIMIT)
            .nsfw(true)
            .status(status)
            .fields([
                "title",
                "num_episodes",
                "average_episode_duration",
                "id",
                "broadcast",
                "start_date",
                "synopsis",
                "alternative_titles",
            ])
            .send()
            .await
        {
            Ok(list) => {
                for aitem in list.data {
                    let anime = aitem.node;
                    if anime.start_date == None {
                        info!("{} has no start date", anime.title);
                        continue;
                    }
                    let start_date = anime
                        .start_date
                        .clone()
                        .unwrap();
                    if start_date.day == None || start_date.month == None {
                        info!("{} has no start date", anime.title);
                        continue;
                    }
                    if anime.broadcast == None {
                        info!("{} has no broadcast time", anime.title);
                        continue;
                    }
                    debug!(
                        "Adding {}",
                        anime
                            .title
                            .clone()
                    );
                    animelist.push(anime.into());
                }
                match list.paging {
                    Some(paging) => {
                        if paging
                            .next
                            .is_none()
                        {
                            break;
                        }
                        offset += LIMIT as u64;
                    },
                    None => break,
                }
            },
            Err(error) => {
                error!("Failed to get {} items: {}", status, error);
                return;
            },
        }
    }
}

pub(crate) async fn update_config(config: &mut config::Config) {
    let server = task::spawn_local(serverfuture());

    let auth: Auth = Auth::new(
        ClientId::new(CLIENTID.to_string()),
        ClientSecret::new(CLIENTSECRET.to_string()),
        RedirectUrl::new(format!("{SCHEME}://{HOST}:{PORT}/")).unwrap(),
    );
    auth.add_scope(Scope::new("write:users".to_string()));
    auth.set_callback(callback)
        .await;
    if config
        .mal_accesskey
        .is_some()
    {
        config
            .mal_accesskey
            .as_ref()
            .unwrap()
            .load(&auth);
    }
    if !auth.is_access_valid() {
        if auth.is_refresh_valid() {
            match auth
                .refresh()
                .await
            {
                Ok(()) => {},
                Err(error) => {
                    log::error!("Failed to refresh token: {}", error);
                    return;
                },
            }
            config.mal_accesskey = Some(MALAccesskey {
                access_token: auth.access_token(),
                refresh_token: auth.refresh_token(),
                access_expiry_time: auth.expires_at(),
                refresh_expiry_time: auth.refresh_expires_at(),
            })
        } else {
            match auth
                .regenerate()
                .await
            {
                Ok(()) => {
                    config.mal_accesskey = Some(MALAccesskey {
                        access_token: auth.access_token(),
                        refresh_token: auth.refresh_token(),
                        access_expiry_time: auth.expires_at(),
                        refresh_expiry_time: auth.refresh_expires_at(),
                    })
                },
                Err(error) => {
                    error!("Failed to update the token: {}", error);
                    return;
                },
            };
        }
    }
    let _ = server.abort();
    let _client = match myanimelist::MalClientBuilder::new()
        .auth(auth)
        .build()
    {
        Ok(client) => client,
        Err(MalClientError::Reqwest(error)) => {
            log::error!("Failed to connect to the API: {}", error);
            return;
        },
        Err(MalClientError::Builder(error)) => {
            log::error!("Failed to build object: {}", error);
            return;
        },
        Err(MalClientError::Token(error)) => {
            log::error!("Broken token: {}", error);
            return;
        },
    };
    let mut animelist = Vec::<Anime>::with_capacity(20);
    debug!("getting all planned animes");
    get_animes(&_client, &mut animelist, WatchStatus::PlanToWatch).await;
    debug!("Getting currently watched animes");
    get_animes(&_client, &mut animelist, WatchStatus::Watching).await;
    debug!("Anime List: {:?}", animelist);
    for anime in animelist {
        if !config
            .anime
            .contains(&anime)
        {
            config
                .anime
                .insert(0, anime);
        }
    }
    config
        .anime
        .sort();
    config
        .anime
        .dedup();
}
